#**************************************************************************
#*                                                                        *
#*                                 OCaml                                  *
#*                                                                        *
#*                 Stephen Dolan, University of Cambridge                 *
#*                                                                        *
#*   Copyright 2016 Stephen Dolan.                                        *
#*                                                                        *
#*   All rights reserved.  This file is distributed under the terms of    *
#*   the GNU Lesser General Public License version 2.1, with the          *
#*   special exception on linking described in the file LICENSE.          *
#*                                                                        *
#**************************************************************************

import gdb

TAGS = {
    246: 'Lazy_tag',
    247: 'Closure_tag',
    248: 'Object_tag',
    249: 'Infix_tag',
    250: 'Forward_tag',
    251: 'Abstract_tag',
    252: 'String_tag',
    253: 'Double_tag',
    254: 'Double_array_tag',
    255: 'Custom_tag'
}

No_scan_tag = 251


debug_tags = {
    0x00: 'Debug_free_minor',
    0x01: 'Debug_free_major',
    0x03: 'Debug_free_shrink',
    0x04: 'Debug_free_truncat',
    0x10: 'Debug_uninit_minor',
    0x11: 'Debug_uninit_major',
    0x15: 'Debug_uninit_align',
    0x85: 'Debug_filler_align'
}

class DoublePrinter:
    def __init__(self, tag, length, p):
        assert tag in ['Double_tag', 'Double_array_tag']
        self.tag = tag
        self.length = length
        self.p = p

    def children(self):
        pass

    def to_string(self):
        return '%s[%d]' % (self.tag, self.length)

class ConstPrinter:
    def __init__(self, rep):
        self.rep = rep
    def to_string(self):
        return self.rep

class BlockPrinter:
    def __init__(self, val):
        if val & 1 == 1:
            self.tag = 1000
            self.tagname = 'I'
            self.val = val
        else:
            self.p = val.cast(val.type.pointer())
            header = (self.p - 1).dereference()
            self.length = int(header >> 10)
            self.gc = int(header & (3 << 8))
            self.tag = int(header & 255)
            self.tagname = TAGS.get(self.tag, 'Block')

    def children(self):
#        if self.tag < No_scan_tag:
#            fields = self.p.cast(gdb.lookup_type('value').pointer())
#            for i in range(self.length):
#                yield '[%d]' % i, (fields + i).dereference()
#        elif self.tagname == 'Double_array_tag':
#            words_per_double = \
#              gdb.lookup_type('double').sizeof / gdb.lookup_type('value').sizeof
#            fields = self.p.cast(gdb.lookup_type('double').pointer())
#            for i in range(int(self.length / words_per_double)):
#                yield '[%d]' % i, (fields + i).dereference()
#
        return []

    def to_string(self):
        if self.tag == 1000:
            # it's an immediate value
            if gdb.lookup_type('value').sizeof == 8:
                debug_mask = 0xff00ffffff00ffff
                debug_val  = 0xD700D7D7D700D6D7
            else:
                debug_mask = 0xff00ffff
                debug_val  = 0xD700D6D7
            n = self.val
            if (n & debug_mask) == debug_val:
                tag = int((n >> 16) & 0xff)
                return debug_tags.get(tag,
                                      "Debug_tag(0x%x)" % int(tag))
            else:
                return "I(%d)" % int(n >> 1)

        # otherwise, it's a block

        if self.tagname == 'Double_tag':
            d = self.p.cast(gdb.lookup_type('double').pointer()).dereference()
            s = '%f, wosize=1' % d
        elif self.tagname == 'String_tag':
            char = gdb.lookup_type('unsigned char')
            val_size = gdb.lookup_type('value').sizeof
            lastbyte = ((self.p + self.length - 1).cast(char.pointer()) + val_size - 1).dereference()
            length_bytes = self.length * val_size - (lastbyte + 1)
            string = (self.p.cast(char.array(length_bytes).pointer()).dereference())
            s = str(string).strip()
        elif self.tagname == 'Infix_tag':
            s = 'offset=%d' % (-self.length)
        elif self.tagname == 'Custom_tag':
            ops = self.p.dereference().cast(gdb.lookup_type('struct custom_operations').pointer())
            s = '%s, wosize=%d' % (str(ops), self.length)
        elif self.tagname == 'Block':
            s = '%d, wosize=%d' % (self.tag,self.length)
        else:
            s = 'wosize=%d' % self.length

        markbits = gdb.lookup_symbol("caml_global_heap_state")[0].value()
        gc = {
            int(markbits['MARKED']): 'MARKED',
            int(markbits['UNMARKED']): 'UNMARKED',
            int(markbits['GARBAGE']): 'GARBAGE',
            (3 << 8): 'NOT_MARKABLE'
        }
        return '%s(%s, %s)' % (self.tagname, s, gc[self.gc])


    def display_hint (self):
        return 'array'



class Fields(gdb.Function):
    def __init__ (self):
        super (Fields, self).__init__ ("F")

    def invoke (self, val):
        assert str(val.type) == 'value'
        p = val.cast(val.type.pointer())
        header = (p - 1).dereference()
        length = int(header >> 10)

        return p.cast(val.type.array(length - 1).pointer()).dereference()


Fields()


def value_printer(val):
    if str(val.type) != 'value':
        return None


    return BlockPrinter(val)

gdb.pretty_printers = [value_printer]

print("These GDB extensions are deprecated. Use tools/gdb.py instead.")



#**************************************************************************
#*                                                                        *
#*                                 OCaml                                  *
#*                                                                        *
#*                 Stephen Dolan, University of Cambridge                 *
#*                                                                        *
#*   Copyright 2016 Stephen Dolan.                                        *
#*                                                                        *
#*   All rights reserved.  This file is distributed under the terms of    *
#*   the GNU Lesser General Public License version 2.1, with the          *
#*   special exception on linking described in the file LICENSE.          *
#*                                                                        *
#**************************************************************************

import gdb

# When running inside GDB, the current directory isn't automatically
# found, so we hack the path to find ocaml.py.
import sys
import os
sys.path.append(os.path.dirname(__file__))

import ocaml

# These three classes (GDBType, GDBValue, GDBTarget) provide a
# generic interface to the debugger, to allow debugger-agnostic code
# in ocaml.py to access debugger and process state.  For a description
# of the required slots and methods, see ocaml.py.

class GDBType:
    def __init__(self, t):
        self._t = t

    def pointer(self):
        return GDBType(self._t.pointer())

    def array(self, size):
        # Amazing mis-feature in the GDB interface: the argument
        # of the `array` method is the inclusive upper index bound,
        # while the lower bound is zero. So we need to pass size-1,
        return GDBType(self._t.array(size-1))

    def size(self):
        return self._t.sizeof

class GDBValue:
    def __init__(self, v, target):
        self._v = v
        self._target = target

    def valid(self):
        # unclear what else this could mean for GDB
        return not (self._v.is_optimized_out)

    def unsigned(self):
        bits = int(self._v)
        if bits < 0:
            bits += (1 << (self._target.word_size * 8))
        return bits

    def signed(self):
        return int(self._v)

    def type(self):
        return GDBType(self._v.type)

    def cast(self, t):
        return GDBValue(self._v.cast(t._t), self._target)

    def value(self):
        return self.cast(self._target._value_type)

    def pointer(self):
        return self.cast(self._target._value_ptr_type)

    def dereference(self):
        return GDBValue(self._v.dereference(), self._target)

    def struct(self):
        return {f.name: GDBValue(self._v[f], self._target)
                for f in self._v.type.fields()}

    def array_size(self):
        range = self._v.type.range()
        return range[1]-range[0]+1

    def sub(self, index):
        return GDBValue(self._v[index], self._target)

    def field(self, index):
        res = ((self._v.cast(self._target._value_ptr_type._t) + index)
               .dereference())
        return GDBValue(res, self._target)

    def field_pointer(self, index):
        return GDBValue(self._v.cast(self._target._value_ptr_type._t) + index,
                        self._target)

    def byte_field(self, index):
        return GDBValue((self._v.cast(self._target._char_ptr_type._t) + index)
                        .dereference(), self._target)

    def double_field(self, index):
        return float((self._v.cast(self._target._double_ptr_type._t) + index)
                     .dereference())

    def string(self, length):
        return (bytes(gdb.selected_inferior().read_memory(self._v, length))
                .decode('UTF-8'))

    def c_string(self):
        return self.cast(self._target._char_ptr_type)._v.string()

    def field_array(self, offset, size):
        ptr = self._v.cast(self._target._value_ptr_type._t) + offset
        field0 = ptr.dereference()
        return field0.cast(field0.type.array(size-1))

    def double_array(self, size):
        return self._v.cast(self._target._double_type.array(size-1)._t)

class GDBTarget:
    def __init__(self):
        self._value_type = GDBType(gdb.lookup_type('value'))
        self._value_ptr_type = self._value_type.pointer()
        self._uintnat_type = GDBType(gdb.lookup_type('uintnat'))
        self._uintnat_ptr_type = self._uintnat_type.pointer()
        self._double_type = GDBType(gdb.lookup_type('double'))
        self._double_ptr_type = self._double_type.pointer()
        self._char_type = GDBType(gdb.lookup_type('char'))
        self._char_ptr_type = self._char_type.pointer()

        self.word_size = self._value_type.size()
        self.double_size = self._double_type.size()

    def global_variable(self, name):
        sym = gdb.lookup_symbol(name, domain=gdb.SYMBOL_VAR_DOMAIN)
        return GDBValue(sym[0].value(), self)

    def type(self, typename):
        return GDBType(gdb.lookup_type(typename))

    def symbol(self, address):
        # Annoyingly GDB doesn't provide a progspace.symbol_of_pc()
        # and gdb doesn't recognise OCaml functions as "functions"
        # for the purposes of progspace.block_of_pc(). So we
        # use a GDB command to get at the symbol.
        text = gdb.execute(f'info symbol 0x{address:x}', to_string=True)
        if not text.startswith('No symbol matches'):
            len = text.find(' in section ')
            if len > 0:
                return text[:len]

    def mapping(self, addr):
        # Annoyingly the progspace.solib_name() and
        # objfile_for_address() functions either aren't reliably
        # present on older versions of GDB, or return unhelpful
        # answers. So we use parse the output of `info proc mapping`.
        # This may be fragile to changes in GDB.
        text = gdb.execute('info proc mappings', to_string=True)
        all_mappings = [m.split() for m in text.split('\n') if '0x' in m]
        mappings = [m for m in all_mappings
                    if int(m[0],0) <= addr < int(m[1],0)]
        if not mappings:
            return
        file_mappings = [m[5] for m in mappings
                         if len(m) > 5
                         and not m[5].startswith('[')]
        # will be surprising if there's more than one of these
        if file_mappings:
            return ', '.join(file_mappings)

    def value(self, v):
        return GDBValue(gdb.Value(v).cast(self._value_type._t),
                        self)

# Object obeying Python's iterator protocol, for iterating through the
# children of a value. This gives us slightly nicer display of block
# values.

class BlockChildren:
    def __init__(self, value):
        self.value = value
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= self.value.num_children:
            raise StopIteration
        element = self.value.child(self.index)
        if isinstance(element, GDBValue):
            element = element._v
        res = (str(self.index), element)
        self.index += 1
        return res

# For pretty-printing values, GDB needs an object with a to_string
# method.  Rather than pushing that into ocaml.Value, we wrap that
# class in a GDB-specific one here.

class ValuePrinter:
    def __init__(self, value):
        target = GDBTarget()
        self._v = ocaml.Value(GDBValue(value, target), target)

    def to_string(self):
        return str(self._v)

    # For pretty-printing block values with children, we
    # need a number of additional methods (which basically
    # delegate to the BlockChildren class above).

    def display_hint(self):
        if self._v.children:
            return 'array'
        else:
            return None

    def children(self):
        return BlockChildren(self._v)

    def num_children(self):
        return self._v.num_children

    def child(self, n):
        return self._v.child(n)

# The actual GDB pretty-printer.

def value_printer(val):
    if str(val.type) != 'value':
        return None
    return ValuePrinter(val)

gdb.pretty_printers = [value_printer]

# Interface to OCaml block finder

class OCamlCommand(gdb.Command):
    "Prefix of all GDB commands for debugging OCaml."
    def __init__(self):
        super(OCamlCommand, self).__init__("ocaml",
                                           gdb.COMMAND_USER, prefix=True)

OCamlCommand()

class OCamlFind(gdb.Command):
    "ocaml find <expr>: report the location of <expr> on the OCaml heap."
    def __init__(self):
        super(OCamlFind, self).__init__("ocaml find", gdb.COMMAND_USER)

    def invoke(self, arg, from_tty):
        self.dont_repeat()
        target = GDBTarget()
        val = ocaml.Value(GDBValue(gdb.parse_and_eval(arg),
                                   target),
                          target)
        ocaml.Finder(target).find(arg, val)

OCamlFind()

# A convenience function $Array which casts a value to an array of values.

class Array(gdb.Function):
    """Turns a Caml value into an array."""
    def __init__ (self):
        super (Array, self).__init__ ("Array")

    def invoke (self, val):
        assert str(val.type) == 'value'
        target = GDBTarget()
        v = ocaml.Value(GDBValue(val, target), target)
        return v.child_array()

Array()

print("OCaml support module loaded. Values of type 'value' will now\n"
      "print as OCaml values, there is a $Array() convenience function,\n"
      "and an 'ocaml' command is available for heap exploration\n"
      "(see 'help ocaml' for more information).")



#**************************************************************************
#*                                                                        *
#*                                 OCaml                                  *
#*                                                                        *
#*                           Nick Barnes, Tarides                         *
#*                                                                        *
#*   Copyright 2024 Tarides.                                              *
#*                                                                        *
#*   All rights reserved.  This file is distributed under the terms of    *
#*   the GNU Lesser General Public License version 2.1, with the          *
#*   special exception on linking described in the file LICENSE.          *
#*                                                                        *
#**************************************************************************

# This file contains any debugger-agnostic code for debugger plugins.
#
# Each debugger front end has three ciasses - targets, types, and
# values, which must provide these slots and methods:
#
# targets:
#
#     word_size: size of a word in 8-bit bytes.
#
#     double_size: the number of 8-bit bytes in a double-precision
#        float (i.e. 8).
#
#     value(address): a debugger "value" representing the given address
#         with type `value`.
#
#     global_variable(name): the value of a named global variable.
#
#     type(name): a named type.
#
#     symbol(address): the symbol name associated with the given address.
#
#     mapping(address): a string describing any file mapping
#         associated with the address, or None. Blocks on the Caml
#         heap do not have an associated file mapping.
#
# types:
#
#     size(): the number of bytes required for this type.
#
#     pointer(): the type of a pointer to this type.
#
#     array(size): the type of an array of given size of this type.
#
# values:
#
#     valid(): False if this value is somehow "invalid" (for example,
#       optimised away.
#
#     type(): The type of this value.
#
#     unsigned(): the value as an unsigned integer.
#
#     signed(): the value as a signed integer.
#
#     cast(t): the value cast to type `t`.
#
#     value(): the value cast to the "value" Caml runtime type.
#
#     pointer(): the value cast to "value*"
#
#     dereference(): the result of dereferencing the value, which must
#       be a pointer.
#
#     array_size(): only used when the value is an array, returns the
#       number of entries in the array.
#
#     sub(index): only used when the value is an array, returns an
#       entry, as a debugger value. TODO: switch to __getitem__ to make
#       this more transparent.
#
#     struct(): a dictionary {slot_name: value} representing the value,
#       which must be a struct. TODO: could use __getattribute__ to make
#       this more transparent.
#
#     field(index): the `index`th field of a value array whose address is
#       in the value, which can have any scalar type. `index` is a Python
#       number which can have any value, including negative.
#
#     field_pointer(index): a pointer to the `index`th field (see above).
#
#     byte_field(index): a byte value, from a byte array.
#
#     double_field(index): a double-precision floating-point value
#       from an array, as a Python float.
#
#     string(length): treating the value as an address, `length` bytes from
#       memory, decoded as UTF-8, as a Python string.
#
#     c_string(): treating the value as an address, returns the NUL-terminated
#       C string at the location as a Python string.
#
#     field_array(offset, size): an array of elements [offset,
#       offset+size), as a debugger-native array.
#
#     double_array(size): an array of double-precision floating point
#       members, as a debugger-native array.

MAX_BLOCK_SLOTS = 8
MAX_STRING_LEN = 80
STRING_SUFFIX = 8
STRING_PREFIX = MAX_STRING_LEN - STRING_SUFFIX - 5
MAX_STRING_SUMMARY = 20
STRING_SUMMARY_PREFIX = 8
STRING_SUMMARY_SUFFIX = MAX_STRING_SUMMARY - STRING_SUMMARY_PREFIX - 5

TAGS = {
    244: 'Forcing',
    245: 'Cont',
    246: 'Lazy',
    247: 'Closure',
    248: 'Object',
    249: 'Infix',
    250: 'Forward',
    251: 'Abstract',
    252: 'String',
    253: 'Double',
    254: 'Double_array',
    255: 'Custom'
}

# specific tag values which we display in particular ways.

TAG_CLOSURE = 247
TAG_INFIX = 249
TAG_STRING = 252
TAG_DOUBLE = 253
TAG_DOUBLE_ARRAY = 254
TAG_CUSTOM = 255

# constants for header word decoding.

HEADER_TAG_BITS = 8
HEADER_TAG_MASK = (1 << HEADER_TAG_BITS) - 1

HEADER_COLOR_BITS = 2
HEADER_COLOR_SHIFT = HEADER_TAG_BITS
HEADER_COLOR_MASK = ((1 << HEADER_COLOR_BITS) - 1) << HEADER_COLOR_SHIFT
NOT_MARKABLE = 3 << HEADER_COLOR_SHIFT

HEADER_WOSIZE_SHIFT = HEADER_TAG_BITS + HEADER_COLOR_BITS

# tags of this or above indicate blocks with no scannable fields
NO_SCAN_TAG = 251

# The debug runtime fills free and uninitialized memory with words:
#
#             D7xx D6D8 on 32-bit platforms
#   D7xx D7D7 D7xx D6D8 in 64-bit platforms
#
# where xx is one of the following 8-bit values, depending on the
# context of the memory word.

DEBUG_TAGS = {
    0x00: 'free minor',
    0x01: 'free major',
    0x03: 'free shrink',
    0x04: 'free truncate', # obsolete
    0x05: 'free unused',
    0x10: 'uninit minor',
    0x11: 'uninit major',
    0x15: 'uninit align',
    0x85: 'filler align',
    0x99: 'pool magic',
}

DEBUG_LOW_BYTES = [0xd8, 0xd6]
DEBUG_OTHER = 0xd7
DEBUG_TAG_BYTES = [2, 6]

def debug_decode(word, word_size):
    """If `word` is a debug padding word, return a string representation of
it. Otherwise, return None. `target` is used for word size."""

    if (word >> (word_size * 8)) not in {0,-1}:
        return
    bytes = [(word >> (i * 8)) & 0xff for i in range(word_size)]
    if bytes[:len(DEBUG_LOW_BYTES)] != DEBUG_LOW_BYTES:
        return
    pads = set(bytes[i]
               for i in range(len(DEBUG_LOW_BYTES), word_size)
               if i not in DEBUG_TAG_BYTES)
    if pads != {DEBUG_OTHER}: # not all pad bytes DEBUG_OTHER
        return
    tags = set(bytes[i] for i in DEBUG_TAG_BYTES if i < word_size)
    if len(tags) != 1: # differing tags on 64-bits
        return
    tag = list(tags)[0] # unique tag byte
    if tag not in DEBUG_TAGS:
        return f'Debug(0x{tag:x}?!)'
    return f'Debug({DEBUG_TAGS[tag]})'

# we show colors as [x], for some character x:

COLOR_SUMMARY = {
    'MARKED': 'm',
    'UNMARKED': 'u',
    'GARBAGE': 'g',
    'NOT MARKABLE': '-',
}

def colors(target):
    """Return a dictionary value -> name of the current GC colors
    (MARKED, UNMARKED, GARBAGE, NOT MARKABLE).
    """

    heapState = target.global_variable('caml_global_heap_state').struct()
    cols = {v.unsigned(): (f, COLOR_SUMMARY[f])
            for f, v in heapState.items()}
    cols[NOT_MARKABLE] = ('NOT MARKABLE', '-')
    return cols

class Value:
    def __init__(self, value, target):
        self._value = value
        self._target = target
        self.children = False
        self.num_children = 0
        self.valid = value.valid()
        if not self.valid:
            return

        self.word = value.signed()
        if self.word == 0:
            self.valid = False
            return

        self.immediate = (self.word & 1) == 1
        if self.immediate:
            return

        self.debug = debug_decode(self.word, target.word_size)
        if self.debug is not None:
            return

        self.pointer = value.pointer()
        self._header = self.pointer.field(-1).unsigned()
        self._wosize = self._header >> HEADER_WOSIZE_SHIFT
        self._tag = self._header & HEADER_TAG_MASK
        self._color_bits = self._header & HEADER_COLOR_MASK
        self.children = True
        self.num_children = self._wosize # overridden for some tags
        if self._tag == TAG_DOUBLE:
            self.children = False
            self.num_children = 0
        elif self._tag == TAG_DOUBLE_ARRAY:
            self.num_children = ((self.num_children * target.word_size)
                                 // target.double_size)
        elif self._tag == TAG_STRING:
            self.children = False
            self.num_children = 0
            byteSize = target.word_size * self._wosize
            lastByte = value.byte_field(byteSize-1).unsigned()
            self._length = byteSize-1-lastByte
            if self._length > 0:
                self._string = value.string(self._length)
            else:
                self._string = ''
        elif self._tag == TAG_CLOSURE:
            # collect code pointers and metadata for all functions
            # in this closure.
            self._functions = []
            # list of (code, arity [, additional code]) tuples.
            self._infix_map = {}
            # map from infix offset to tuple
            arity_shift = target.word_size * 8 - 8
            closinfo = value.field(1).signed()
            self._start_env = (closinfo & ((1 << arity_shift) - 1)) >> 1
            self.num_children = self._wosize - self._start_env
            block = 0
            while block < self._start_env:
                code = value.field(block).unsigned()
                closinfo = value.field(block+1).unsigned()
                arity = closinfo >> arity_shift
                if (arity == 0) or (arity == 1):
                    fn = (code, arity)
                    bump = 0
                else: # higher arity, so code is curry/tuplify
                    true_code = value.field(block+2).unsigned()
                    fn = (true_code, arity, code)
                    bump = 1
                self._functions.append(fn)
                self._infix_map[block] = fn
                block += 3 +bump # code, closinfo, [extra code], [infix header]
        elif self._tag == TAG_INFIX:
            self._container = Value(value.field_pointer(-self._wosize), target)
            self.num_children = 0
            self.children = False
        elif self._tag == TAG_CUSTOM:
            ptr_type = target.type('struct custom_operations').pointer()
            self._ops = value.field(0).cast(ptr_type).dereference().struct()
            self._id = self._ops['identifier'].c_string()
            self.children = False
            self.num_children = 0

    def tag_part(self):
        if self._tag in TAGS:
            return f'{TAGS[self._tag]}'
        elif self._tag == 0:
            return ''
        else:
            return f't{self._tag}'

    def infix_sym(self):
        cont = self._container
        sym = f'+{self._wosize}'
        # try to find symbol in infix map of container
        if cont._tag == TAG_CLOSURE: # always true!
            if self._wosize in cont._infix_map:
                code = cont._infix_map[self._wosize][0]
                sym = self._target.symbol(code)
                if sym is None:
                    sym = f'0x{code:x}'
        return sym

    def code_sym(self, t):
        code = t[0]
        sym = self._target.symbol(code)
        if sym is None:
            sym = f'0x{code:x}'
        if len(t) == 2:
            return sym
        else:
            return f'{sym}({self._target.symbol(t[2])})'

    def closure_syms(self):
        return [self.code_sym(t) for t in self._functions]

    def array_contents(self, short=False):
        if self._tag == TAG_DOUBLE_ARRAY:
            if self.num_children <= MAX_BLOCK_SLOTS:
                return [str(self._value.double_field(i))
                        for i in range(self.num_children)]
            return ([str(self._value.double_field(i))
                     for i in range(MAX_BLOCK_SLOTS-2)]
                    + ['...',
                       str(self._value.double_field(
                           self.num_children - 1))])

        if self.num_children < MAX_BLOCK_SLOTS:
            return [self.field_summary(i, short)
                    for i in range(self.num_children)]
        else:
            return ([self.field_summary(i, short)
                     for i in range(MAX_BLOCK_SLOTS-2)]
                    + ['...',
                       self.field_summary(self.num_children - 1, short)])

    def summary(self, short=False):
        """Return a short value summary string, suitable for display in a
        larger aggregate. If `short` then summarise the summary.
        """
        if not self.valid:
            return '[invalid]'
        if self.immediate:
            return f'{self.word // 2}'
        if self.debug is not None:
            return self.debug

        if self._tag == TAG_DOUBLE:
            return str(self._value.double_field(0))
        elif self._tag == TAG_STRING:
            if self._length > MAX_STRING_SUMMARY:
                return (repr(self._string[:STRING_SUMMARY_PREFIX])
                        + '...'
                        + repr(self._string[-STRING_SUMMARY_SUFFIX:])
                        + f'<{self._length}>')
            return repr(self._string)
        elif self._tag == TAG_INFIX:
            sym = self.infix_sym()
            return f'infix({sym}) in ' + self._container.summary(short=True)
        elif self._tag == TAG_CLOSURE:
            syms = self.closure_syms()
            if len(syms) > 1:
                sym = f'{syms[0]}, +{len(syms)-1}'
            else:
                sym = syms[0]
            return f'closure({sym})<{self.num_children}>'
        elif self._tag == TAG_CUSTOM:
            return f"custom {self._id}<{self._wosize}>"

        tag_part = self.tag_part()

        if short:
            if not tag_part:
                tag_part = 't0'
            return f'<{tag_part}:{self.num_children}>'

        if tag_part:
            tag_part += ':'

        contents = self.array_contents(short=True)

        return (f'({tag_part}' + ', '.join(contents) + ')')


    def field_summary(self, index, short=False):
        return (Value(self._value.field(index), self._target).
                summary(short))

    def __str__(self):
        if not self.valid:
            return '[invalid]'
        if self.immediate:
            return f'caml:{self.word // 2}'
        if self.debug is not None:
            return f'Caml:{self.debug}'

        color_char = colors(self._target).get(self._color_bits,
                                         f'BAD COLOR {self._color_bits}')[1]
        prefix = f'caml({color_char}):'

        if self._tag == TAG_DOUBLE:
            val = str(self._value.double_field(0))
            return f'{prefix}{val}'
        elif self._tag == TAG_STRING:
            if self._length > MAX_STRING_LEN:
                s = (repr(self._string[:STRING_PREFIX])
                     + '...' + repr(self._string[-STRING_SUFFIX:]))
            else:
                s = repr(self._string)
            return (f'{prefix}{s}<{self._length}>')
        elif self._tag == TAG_INFIX:
            sym = self.infix_sym()
            return (f'{prefix}infix({sym}) in'
                    + f' 0x{self._container._value.unsigned():x} '
                    + self._container.summary())
        elif self._tag == TAG_CLOSURE:
            syms = ', '.join(self.closure_syms())
            return (f'{prefix}closure({syms})'
                    + f' arity {self._functions[0][1]} ('
                    + ', '.join(self.field_summary(i + self._start_env)
                                for i in range(self.num_children))
                    + ')')
        elif self._tag == TAG_CUSTOM:
            return (f"{prefix}custom {self._id}"
                    f"<{self._wosize}>")

        tag_part = self.tag_part()
        if self._tag != 0:
            tag_part += ': '
        suffix = ('' if self.num_children <= MAX_BLOCK_SLOTS
                  else f'<{self.num_children}>')

        contents = self.array_contents()

        return (f'{prefix}({tag_part}'
                + ', '.join(contents)
                + f'){suffix}')

    # Useful in GDB and maybe one day in LLDB too.

    def child(self, index):
        if (not self.children) or index < 0 or index >= self.num_children:
            return
        if self._tag == TAG_DOUBLE_ARRAY:
            return self._value.double_field(index)
        elif self._tag == TAG_CLOSURE:
            return self._value.field(index + self._start_env).value()
        else:
            return self._value.field(index).value()

    # Useful in GDB and maybe one day in LLDB too.

    def child_array(self):
        """If the value is a block which can be regarded as an array,
        return the array as a debugger-native value."""
        if (not self.children):
            return
        if self._tag == TAG_DOUBLE_ARRAY:
            return self._value.double_array(self.num_children)
        elif self._tag == TAG_CLOSURE:
            return self._value.field_array(self._start_env, self.num_children)
        else:
            return self._value.field_array(0, self.num_children)

POOL_WSIZE = 4096

class Finder:
    def __init__(self, target):
        self._sizeclasses = None
        self._wsize_sizeclass = None
        self._target = target
        self.debug = False

    def sizeclasses(self):
        if self._sizeclasses is None:
            pool_freelist = (self._target.global_variable('pool_freelist').
                             struct())
            self._sizeclasses = (
                pool_freelist['global_avail_pools'].type().size() //
                pool_freelist['global_avail_pools'].sub(0).type().size())
        return self._sizeclasses

    def wsize_sizeclass(self, sz):
        if self._wsize_sizeclass is None:
            self._wsize_sizeclass = (self._target.
                                     global_variable('wsize_sizeclass'))
        return self._wsize_sizeclass.sub(sz).unsigned()

    def _log(self, *args):
        if self.debug:
            print(*args)

    def _found(self, where):
        if self.debug:
            print(f"FOUND 0x{self.address:x} {where}")
        self.found.append(where)
        self.keep_going = self.debug

    def search_pool_list(self, description, pool_list):
        "Search a single pool list for `self.address`."
        count = 0
        while self.keep_going and pool_list.unsigned():
            count += 1
            base = pool_list.unsigned()
            limit = base + POOL_WSIZE * self._target.word_size
            if base < self.address < limit:
                self._found(f"{description}: pool 0x{base:x}-0x{limit:x}")
            pool_list = pool_list.dereference().struct()['next']
        self._log(f"    searched {count} pools of {description}")

    def search_pools(self, description, pools):
        "Search an array `pool *pools[NUM_SIZECLASSES]` for `self.address`."
        self._log(f"  searching {description} pools")
        for i in range(self.sizeclasses()):
            pool_list = pools.sub(i)
            if pool_list.unsigned() == 0:
                continue
            self.search_pool_list(f"{description} "
                                  f"wsize={self.wsize_sizeclass(i)}",
                                  pool_list)
            if not self.keep_going:
                break

    def search_large(self, description, large_list):
        "Search a `large_alloc *` linked list for `self.address`."
        if large_list.unsigned() == 0:
            return
        count = 0
        while self.keep_going and large_list.unsigned():
            count += 1
            base = large_list.unsigned()
            block = base + large_list.dereference().type().size()
            val = self._target.value(block + self._target.word_size)
            val_ptr = val.cast(val.type().pointer())
            oval = Value(val_ptr, self._target)
            limit = block + (oval._wosize + 1) * self._target.word_size
            if base < self.address < limit:
                self._found(f"{description} large 0x{block:x}-0x{limit:x}")
            large_list = large_list.dereference().struct()['next']
        self._log(f"  searched {count} large blocks of {description}")

    def search_heap(self, description, heap_state_p):
        "Searches a single `struct caml_heap_state *` for self.address."
        heap_state = heap_state_p.dereference().struct()
        if self.keep_going:
            self.search_pools(f"{description} avail",
                              heap_state['avail_pools'])
        if self.keep_going:
            self.search_pools(f"{description} full",
                              heap_state['full_pools'])
        if self.keep_going:
            self.search_pools(f"{description} unswept avail",
                              heap_state['unswept_avail_pools'])
        if self.keep_going:
            self.search_pools(f"{description} unswept full",
                              heap_state['unswept_full_pools'])
        if self.keep_going:
            self.search_large(f"{description}",
                              heap_state['swept_large'])
        if self.keep_going:
            self.search_large(f"{description} unswept",
                              heap_state['unswept_large'])

    def search_domain(self, index, dom_state_p):
        "Search a single domain's heap for `self.address`."
        dom_state = dom_state_p.dereference().struct()
        young_start = dom_state['young_start'].unsigned()
        young_end = dom_state['young_end'].unsigned()
        description = f"domain {index}"
        self._log(f"searching {description}")
        if self.keep_going and (young_start <= self.address <= young_end):
                self._found(f"{description} minor heap "
                            f"0x{young_start:x}-0x{young_end:x}")
        if self.keep_going:
            self.search_heap(description, dom_state['shared_heap'])

    def find(self, expr, val):
        if not val.valid:
            print(f"{expr} not a valid expression")
            return
        if val.immediate:
            print(f"{expr} is immediate: {str(val)}")
            return
        if val.debug is not None:
            print(f"{expr} is a debug padding value: {val.debug}")
            return

        self.address = val.pointer.unsigned()
        mapping = self._target.mapping(self.address)
        if mapping:
            print(f"{expr} {str(val)} is from {mapping}, "
                  "not the heap.")
            return

        self.found = []
        self.keep_going = True

        # Search per-domain heaps.
        all_domains = self._target.global_variable('all_domains')
        Max_domains = all_domains.array_size()
        self._log(f"{Max_domains} domains.")
        for i in range(Max_domains):
            dom = all_domains.sub(i).struct()
            dom_state_p = dom['state']
            if dom_state_p.unsigned() == 0: # null pointer: no domain
                continue
            self.search_domain(i, dom_state_p)
            if not self.keep_going:
                break

        # Global (orphaned) heap
        pool_freelist = self._target.global_variable('pool_freelist').struct()
        if self.keep_going:
            self.search_pools('global avail',
                              pool_freelist['global_avail_pools'])
        if self.keep_going:
            self.search_pools('global full',
                              pool_freelist['global_full_pools'])
        if self.keep_going:
            self.search_large("global",
                              pool_freelist['global_large'])

        if self.found:
            print(f"{expr} {str(val)}: 0x{self.address:x} found:")
            for where in self.found:
                print(f"  {where}")
        else:
            print(f"{expr} {str(val)} not found on heap")

        self.debug = True



#**************************************************************************
#*                                                                        *
#*                                 OCaml                                  *
#*                                                                        *
#*                           Nick Barnes, Tarides                         *
#*                                                                        *
#*   Copyright 2024 Tarides.                                              *
#*                                                                        *
#*   All rights reserved.  This file is distributed under the terms of    *
#*   the GNU Lesser General Public License version 2.1, with the          *
#*   special exception on linking described in the file LICENSE.          *
#*                                                                        *
#**************************************************************************

import lldb
import ocaml

def __lldb_init_module(d, _internal_dict):
    global debugger
    debugger = d
    d.HandleCommand('type summary add --python-function '
                    f'{__name__}.show_value value')
    d.HandleCommand('command container add '
                    '-h "OCaml runtime debugging utilities" '
                    'ocaml')
    d.HandleCommand('command script add --class '
                    f'{__name__}.OCamlFind '
                    'ocaml find'
                    )
    print("OCaml support module loaded. Values of type 'value' will now\n"
          "print as OCaml values, and an 'ocaml' command is available for\n"
          "heap exploration (see 'help ocaml' for more information).")

    # Synthetic Child providers don't seem so useful
    # d.HandleCommand("type synthetic add value --python-class "
    #                f"{__name__}.ChildProvider")

class OCamlFind:
    def __init__(self, debugger, internal_dict):
        super()

    def __call__(self, debugger, expr, exe_ctx, result):
        target = get_target()
        finder = ocaml.Finder(target)
        val = ocaml.Value(LLDBValue(exe_ctx.GetFrame().EvaluateExpression(expr),
                                    target),
                          target)
        finder.find(expr, val)

    def get_short_help(self):
        return "Describe the location of the given OCaml value in the heap."
    def get_long_help(self):
        return "Describe the location of the given OCaml value in the heap."


# These three classes (LLDBType, LLDBValue, LLDBTarget) provide a
# generic interface to the debugger, to allow debugger-agnostic code
# in ocaml.py to access debugger and process state.  For a description
# of the required slots and methods, see ocaml.py.

class LLDBType:
    def __init__(self, t):
        self._t = t

    def pointer(self):
        return LLDBType(self._t.GetPointerType())

    def array(self, size):
        return LLDBType(self._t.GetArrayType(size))

    def size(self):
        return self._t.size


class LLDBValue:
    def __init__(self, v, target):
        self._v = v
        self._target = target

    def valid(self):
        return self._v.IsValid()

    def unsigned(self):
        return self._v.unsigned

    def signed(self):
        return self._v.signed

    def type(self):
        return LLDBType(self._v.type)

    def cast(self, t):
        return LLDBValue(self._v.Cast(t._t), self._target)

    def value(self):
        return self.cast(self._target._value_type)

    def pointer(self):
        return self.cast(self._target._value_ptr_type)

    def dereference(self):
        return LLDBValue(self._v.Dereference(), self._target)

    def struct(self):
        t = self._v.GetType()
        fields = t.GetNumberOfFields()
        return {member.name:
                    LLDBValue(self._v.GetChildMemberWithName(member.name),
                              self._target)
                for i in range(t.GetNumberOfFields())
                if (member := t.GetFieldAtIndex(i))}

    def array_size(self):
        return self._v.GetNumChildren()

    def sub(self, index):
        return LLDBValue(self._v.GetChildAtIndex(index,
                                                 lldb.eNoDynamicValues,
                                                 False),
                         self._target)

    def field(self, index):
        address = self.unsigned() + index * self._target.word_size
        return self._target._create_value(f'[{index}]', address,
                                          self._target._value_type)

    def field_pointer(self, index):
        return LLDBValue(self.field(index)._v.AddressOf(), self._target)

    def byte_field(self, index):
        ptr = self._v.Cast(self._target._char_ptr_type._t)
        return LLDBValue(ptr.GetChildAtIndex(index,
                                             lldb.eNoDynamicValues,
                                             True),
                         self._target)

    # converts to/from a string; could use GetValueAsUnsigned and the
    # `struct` module instead.
    def double_field(self, index):
        ptr = self._v.Cast(self._target._double_ptr_type._t)
        return float(ptr.GetChildAtIndex(index,
                                         lldb.eNoDynamicValues,
                                         True)
                     .GetValue())

    def string(self, length):
        return self._target._memory(self.unsigned(), length).decode('UTF-8')

    def c_string(self):
        return self._target._c_string(self.unsigned())

    def field_array(self, offset, size):
        address = self.unsigned() + offset * self._target.word_size
        val = self._target._create_value('[]', address,
                                         self._target._value_type.array(size))

    def double_array(self, size):
        return self._v.Cast(self._target._double_type.array(size)._t)

class LLDBTarget:
    def __init__(self, target):
        self._target = target
        self._value_type = LLDBType(target.FindFirstType("value"))
        self._value_ptr_type = self._value_type.pointer()
        self._double_type = LLDBType(target.FindFirstType("double"))
        self._double_ptr_type = self._double_type.pointer()
        self._char_type = LLDBType(target.FindFirstType("char"))
        self._char_ptr_type = self._char_type.pointer()

        self.word_size = self._value_type.size()
        self.double_size = self._double_type.size()

    def global_variable(self, name):
        return LLDBValue(self._target.FindFirstGlobalVariable(name),
                         self)

    def type(self, typename):
        return LLDBType(self._target.FindFirstType(typename))

    def symbol(self, address):
        addr = lldb.SBAddress(address, self._target)
        if addr.IsValid() and addr.symbol and addr.symbol.name:
            return addr.symbol.name

    def mapping(self, addr):
        address = self._address(addr)
        section = address.GetSection()
        module = address.GetModule()
        if section.name:
            return f"{str(module.file)}:{section.name}"


    def value(self, v):
        return LLDBValue(self._target.EvaluateExpression(f"((value){v})"),
                         self)

    ## These methods are only used by methods inside LLDBValue and/or LLDBType

    def _address(self, addr):
        return lldb.SBAddress(addr, self._target)

    def _memory(self, addr, len):
        return self._target.process.ReadMemory(addr, len, lldb.SBError())

    def _c_string(self, addr):
        return self._target.process.ReadCStringFromMemory(addr, 256,
                                                          lldb.SBError())

    def _create_value(self, name, addr, ty):
        val = self._target.CreateValueFromAddress(name,
                                                  self._address(addr),
                                                  ty._t)
        return LLDBValue(val, self)

targets = {}

def get_target():
    target = debugger.GetSelectedTarget()
    # TODO: SBTarget not hashable, but we want something hashable to
    # uniquely represent it, for advanced cases in which we end up
    # with (for instance) more than one process in the same debugging
    # session.
    key = (target.triple, target.process.id)
    if key not in targets:
        targets[key] = LLDBTarget(target)
    return targets[key]

def show_value(value, _internal_dict, options):
    target = get_target()
    return str(ocaml.Value(LLDBValue(value, target), target))

# A class like this for aggregate values should let LLDB show members
# in a natural way, but I can't get it to work.

class ChildProvider:
    def __init__(self, value, internal_dict):
        self.value = value
        self.ocaml = None

    def _update(self):
        target = get_target()
        self.ocaml = ocaml.Value(LLDBValue(self.value, target),
                                 target)

    def _ensure(self):
        if self.ocaml is None:
            self._update()

    def has_children(self):
        """Return True if the value might have children,
        False if definitely not."""
        self._ensure()
        print(f"has_children called on {self.ocaml}")
        return self.ocaml.children

    def num_children(self):
        "Return the number of children."
        self._ensure()
        print(f"num_children called on {self.ocaml}")
        return self.ocaml.num_children

    def get_child_index(self, name):
        """Return the index of a child, given part of an expression
        identifying a child. Evidently used by LLDB parser.
        """
        # this is pretty mysterious
        self._ensure()
        print(f"get_child_index called on {self.ocaml}")
        return int(name.lstrip('[').rstrip(']'))

    def get_child_at_index(self, index):
        self._ensure()
        return self.ocaml.child(index)

    def update(self):
        """Update the internal state whenever the state of the variables
        in LLDB changes. Invoked before any other method in the interface."""
        self._update()
        print(f"update called on {self.ocaml}")

    def get_value(self):
        self._update()
        print(f"get_value called on {self.ocaml}")
        if self.ocaml.children:
            return self.value.cast(
                self.value.type.GetArrayType(self.ocaml.num_children))
        else:
            return self.value
